#include <cstdlib>
#include <cstdio>

#define hchk(a,b) if((a) == NULL) printf("H: %s\n", (b));
#define dchk(a,b) if((a) != cudaSuccess) printf("S: @%s: %s\n", (b), cudaGetErrorString(cudaGetLastError()));

const long side = 4;
const long items = side*side*side;

/**
 * \brief Init an array as the side of a cube
 * \param x A pointer to the array to init
 * \param side The length of the side (no. of items)
 */
void cube(double* x, const long side)
{
	int i, j, k;
	double start = -(side >> 1);
	for(i=0; i<side; ++i)
	{
		for(j=0; j<side; ++j)
		{
			for(k=0; k<side; ++k)
			{
				x[i*side*side+j*side+k] = start + k;
			}
		}
	}
}

/**
 * \brief Display two arrays of properties
 * \param x A pointer to the array with the first property
 * \param f_x A pointer to the array with the second property
 * \param side The length of the side (no. of items)
 */
void display(double* x, double* f_x, const long side)
{
	int i;
	for(i=0; i<side*side*side; ++i)
	{
		printf("%2.0f  ", x[i]);
		if(i%side == side-1) printf("\n");
		if(i%(side*side) == (side*side-1)) printf("\n\n");
	}
	for(i=0; i<side*side*side; ++i)
	{
		printf("%6.3f ", f_x[i]);
		if(i%side == side-1) printf("\n");
		if(i%(side*side) == (side*side-1)) printf("\n\n");
	}
}

extern __shared__ char local[];
__global__ void kernel(double* rlx, double* rfx)
{
	const long half_loop = blockDim.x;
	const long full_loop = half_loop << 1;
	const long prec_size = sizeof(double);
	const long this_pos = threadIdx.x;
	
	double* llx = (double*)&local[0];
	double* lfx = (double*)&local[full_loop*prec_size];
	
	llx[this_pos] = rlx[this_pos];
	llx[half_loop+this_pos] = rlx[half_loop+this_pos];
	lfx[this_pos] = 0;
	lfx[half_loop+this_pos] = 0;
	
	long item_pos = this_pos+1;
	for(long crosses=0; crosses < half_loop; ++crosses)
	{
		if(item_pos == full_loop) item_pos = 0;
		__syncthreads();
		lfx[this_pos] -= llx[this_pos]-llx[item_pos];
		lfx[item_pos] += llx[this_pos]-llx[item_pos];
		++item_pos;
	}
	__syncthreads();
	rfx[this_pos] = lfx[this_pos];
	rfx[half_loop+this_pos] = lfx[half_loop+this_pos];
}

int main()
{
	// SET WHERE THE DEVICE //
	dchk(cudaSetDevice(1), "setting device");
	
	// ALLOCATE MEMORY //
	size_t items_prop_bytes = items*sizeof(double);
	double* hlx = (double*)malloc(items_prop_bytes);
	double* hfx = (double*)malloc(items_prop_bytes);
	hchk(hlx, "null hlx");
	hchk(hfx, "null hfx");
	
	double* dlx;
	double* dfx;
	dchk(cudaMalloc((void**)&dlx ,items_prop_bytes), "allocating dlx");
	dchk(cudaMalloc((void**)&dfx ,items_prop_bytes), "allocating dfx");
	
	// INIT SPACE //
	cube(hlx, side);
	
	// COPY MEMORY HOST -> DEVICE //
	dchk(cudaMemcpy(dlx, hlx, items_prop_bytes, cudaMemcpyHostToDevice), "copying hlx -> dlx");
	dchk(cudaMemset(dfx, 0, items_prop_bytes), "zeroing dfx");
	
	// RUN KERNEL //
	kernel<<<1, items/2, items*2*sizeof(double)>>>(dlx, dfx);
	
	// COPY MEMORY DEVICE -> HOST //
	dchk(cudaMemcpy(hlx, dlx, items_prop_bytes, cudaMemcpyDeviceToHost), "copying dlx -> hlx");
	dchk(cudaMemcpy(hfx, dfx, items_prop_bytes, cudaMemcpyDeviceToHost), "copying dfx -> hfx");
	
	// PRINT VALUES //
	display(hlx, hfx, side);
	
	// FREE MEMORY //
	dchk(cudaFree(dlx), "freeing dlx");
	dchk(cudaFree(dfx), "freeing dfx");
	
	free(hlx);
	free(hfx);
}